@@ -29,6 +29,7 @@ import (
 	"cuelang.org/go/cue"
 	"cuelang.org/go/cue/errors"
 	"cuelang.org/go/cue/token"
+	"cuelang.org/go/internal/core/adt"
 	itask "cuelang.org/go/internal/task"
 	"cuelang.org/go/internal/value"
 	_ "cuelang.org/go/pkg/tool/cli" // Register tasks
@@ -42,7 +43,7 @@ import (
 const commandSection = "command"
 
 func lookupString(obj cue.Value, key, def string) string {
-	str, err := obj.Lookup(key).String()
+	str, err := obj.LookupPath(cue.MakePath(cue.Str(key))).String()
 	if err == nil {
 		def = str
 	}
@@ -92,17 +93,22 @@ func customCommand(c *Command, typ, name string, tools *cue.Instance) (*cobra.Co
 
 	// Ensure there is at least one tool file.
 	// TODO: remove this block to allow commands to be defined in any file.
-outer:
 	for _, v := range []cue.Value{tools.Lookup(typ), o} {
 		_, w := value.ToInternal(v)
-		for _, c := range w.Conjuncts {
+		hasToolFile := false
+		w.VisitLeafConjuncts(func(c adt.Conjunct) bool {
 			src := c.Source()
 			if src == nil {
-				continue
+				return true
 			}
 			if strings.HasSuffix(src.Pos().Filename(), "_tool.cue") {
-				break outer
+				hasToolFile = true
+				return false
 			}
+			return true
+		})
+		if hasToolFile {
+			break
 		}
 		if err := v.Err(); err != nil {
 			return nil, err
@@ -130,20 +136,22 @@ outer:
 		Use:   usage,
 		Short: lookupString(o, "$short", short),
 		Long:  lookupString(o, "$long", long),
-		RunE: mkRunE(c, func(cmd *Command, args []string) error {
+		// Note that we don't use mkRunE here, as the parent func is already wrapped by
+		// another mkRunE call, and all Command initialization has already happened.
+		RunE: func(cmd *cobra.Command, args []string) error {
 			// TODO:
 			// - parse flags and env vars
 			// - constrain current config with config section
 
-			return doTasks(cmd, typ, name, tools)
-		}),
+			return doTasks(c, name, tools)
+		},
 	}
 
 	// TODO: implement var/flag handling.
 	return sub, nil
 }
 
-func doTasks(cmd *Command, typ, command string, root *cue.Instance) error {
+func doTasks(cmd *Command, command string, root *cue.Instance) error {
 	cfg := &flow.Config{
 		Root:           cue.MakePath(cue.Str(commandSection), cue.Str(command)),
 		InferTasks:     true,
@@ -152,10 +160,7 @@ func doTasks(cmd *Command, typ, command string, root *cue.Instance) error {
 
 	c := flow.New(cfg, root, newTaskFunc(cmd))
 
-	err := c.Run(backgroundContext())
-	exitOnErr(cmd, err, true)
-
-	return err
+	return c.Run(backgroundContext())
 }
 
 // func (r *customRunner) tagReference(t *task, ref cue.Value) error {
@@ -275,7 +280,7 @@ var testServerOnce = sync.OnceValue(func() string {
 	s := httptest.NewServer(http.HandlerFunc(
 		func(w http.ResponseWriter, req *http.Request) {
 			data, _ := io.ReadAll(req.Body)
-			d := map[string]interface{}{
+			d := map[string]string{
 				"data": string(data),
 				"when": "now",
 			}
