@@ -87,6 +87,7 @@ removal.
 	}
 
 	addOutFlags(cmd.Flags(), false)
+	cmd.Flags().BoolP(string(flagDryRun), "n", false, "only run simulation")
 
 	return cmd
 }
@@ -100,7 +101,10 @@ func runTrim(cmd *Command, args []string) error {
 	if binst == nil {
 		return nil
 	}
-	instances := buildInstances(cmd, binst, false)
+	instances, err := buildInstances(cmd, binst, false)
+	if err != nil {
+		return err
+	}
 
 	dst := flagOutFile.String(cmd)
 	if dst != "" && dst != "-" && !flagForce.Bool(cmd) {
@@ -130,7 +134,10 @@ func runTrim(cmd *Command, args []string) error {
 
 	cfg := *defCfg.loadCfg
 	cfg.Overlay = overlay
-	tinsts := buildInstances(cmd, load.Instances(args, &cfg), false)
+	tinsts, err := buildInstances(cmd, load.Instances(args, &cfg), false)
+	if err != nil {
+		return err
+	}
 	if len(tinsts) != len(binst) {
 		return errors.New("unexpected number of new instances")
 	}
@@ -141,12 +148,12 @@ func runTrim(cmd *Command, args []string) error {
 				diff.Print(os.Stdout, script)
 				fmt.Println("Aborting trim, output differs after trimming. This is a bug! Use -i to force trim.")
 				fmt.Println("You can file a bug here: https://cuelang.org/issues/new?assignees=&labels=NeedsInvestigation&template=bug_report.md&title=")
-				os.Exit(1)
+				return ErrPrintedError
 			}
 		}
 	}
 
-	if flagDryrun.Bool(cmd) {
+	if flagDryRun.Bool(cmd) {
 		return nil
 	}
 
