@@ -87,16 +87,16 @@ func nestDepth(f *ast.Field) int {
 
 // TODO: be more accurate and move to astutil
 func hasDocComments(d ast.Decl) bool {
-	if len(d.Comments()) > 0 {
+	if len(ast.Comments(d)) > 0 {
 		return true
 	}
 	switch x := d.(type) {
 	case *ast.Field:
-		return len(x.Label.Comments()) > 0
+		return len(ast.Comments(x.Label)) > 0
 	case *ast.Alias:
-		return len(x.Ident.Comments()) > 0
+		return len(ast.Comments(x.Ident)) > 0
 	case *ast.LetClause:
-		return len(x.Ident.Comments()) > 0
+		return len(ast.Comments(x.Ident)) > 0
 	}
 	return false
 }
@@ -104,7 +104,7 @@ func hasDocComments(d ast.Decl) bool {
 func (f *formatter) walkDeclList(list []ast.Decl) {
 	f.before(nil)
 	d := 0
-	hasEllipsis := false
+	var ellipsis ast.Decl
 	for i, x := range list {
 		if i > 0 {
 			f.print(declcomma)
@@ -128,7 +128,7 @@ func (f *formatter) walkDeclList(list []ast.Decl) {
 			}
 		}
 		if f.printer.cfg.simplify && internal.IsEllipsis(x) {
-			hasEllipsis = true
+			ellipsis = x
 			continue
 		}
 		f.decl(x)
@@ -156,8 +156,11 @@ func (f *formatter) walkDeclList(list []ast.Decl) {
 		}
 		f.print(f.current.parentSep)
 	}
-	if hasEllipsis {
-		f.decl(&ast.Ellipsis{})
+	if ellipsis != nil {
+		// ensure that comments associated with the original ellipsis are preserved
+		n := &ast.Ellipsis{}
+		ast.SetComments(n, ast.Comments(ellipsis))
+		f.decl(n)
 		f.print(f.current.parentSep)
 	}
 	f.after(nil)
@@ -175,9 +178,12 @@ func (f *formatter) walkSpecList(list []*ast.ImportSpec) {
 
 func (f *formatter) walkClauseList(list []ast.Clause, ws whiteSpace) {
 	f.before(nil)
-	for _, x := range list {
+	for i, x := range list {
 		f.before(x)
-		f.print(ws)
+		// Only print the whitespace between the clauses.
+		if i > 0 {
+			f.print(ws)
+		}
 		f.clause(x)
 		f.after(x)
 	}
@@ -188,6 +194,26 @@ func (f *formatter) walkListElems(list []ast.Expr) {
 	f.before(nil)
 	for _, x := range list {
 		f.before(x)
+
+		// This is a hack to ensure that comments are printed correctly in lists.
+		// A comment must be printed after each element in a list, but we can't
+		// print a comma at the end of a comment because it will be considered
+		// part of the comment and ignored.
+		// To fix this we collect all comments that appear after the element,
+		// and only handle them after it's formatted.
+		var commentsAfter []*ast.CommentGroup
+		splitComments := x.Pos().IsValid()
+		if splitComments {
+			for _, cg := range ast.Comments(x) {
+				if x.Pos().Before(cg.Pos()) {
+					commentsAfter = append(commentsAfter, cg)
+				}
+			}
+		}
+
+		if splitComments {
+			f.current.cg = nil
+		}
 		switch n := x.(type) {
 		case *ast.Comprehension:
 			f.walkClauseList(n.Clauses, blank)
@@ -208,6 +234,10 @@ func (f *formatter) walkListElems(list []ast.Expr) {
 			f.exprRaw(n, token.LowestPrec, 1)
 		}
 		f.print(comma, blank)
+
+		if splitComments {
+			f.current.cg = commentsAfter
+		}
 		f.after(x)
 	}
 	f.after(nil)
@@ -235,7 +265,7 @@ func (f *formatter) inlineField(n *ast.Field) *ast.Field {
 	regular := internal.IsRegularField(n)
 	// shortcut single-element structs.
 	// If the label has a valid position, we assume that an unspecified
-	// Lbrace signals the intend to collapse fields.
+	// Lbrace signals the intent to collapse fields.
 	if !n.Label.Pos().IsValid() && !(f.printer.cfg.simplify && regular) {
 		return nil
 	}
@@ -284,6 +314,7 @@ func (f *formatter) decl(decl ast.Decl) {
 		} else {
 			f.print(blank, nooverride, n.Token)
 		}
+		f.visitComments(f.current.pos)
 
 		if mem := f.inlineField(n); mem != nil {
 			switch {
@@ -304,7 +335,7 @@ func (f *formatter) decl(decl ast.Decl) {
 
 		nextFF := f.nextNeedsFormfeed(n.Value)
 		tab := vtab
-		if nextFF {
+		if nextFF || f.prevLbraceOnLine {
 			tab = blank
 		}
 
@@ -380,7 +411,7 @@ func (f *formatter) decl(decl ast.Decl) {
 			f.print(formfeed)
 		}
 		f.expr(n.Expr)
-		f.print(newline, noblank)
+		f.print(newline)
 
 	case *ast.Attribute:
 		f.print(n.At, n)
@@ -431,10 +462,16 @@ func (f *formatter) nextNeedsFormfeed(n ast.Expr) bool {
 		return strings.IndexByte(x.Value, '\n') >= 0
 	case *ast.ListLit:
 		return true
+	case *ast.ParenExpr:
+		return f.nextNeedsFormfeed(x.X)
 	case *ast.UnaryExpr:
 		return f.nextNeedsFormfeed(x.X)
 	case *ast.BinaryExpr:
 		return f.nextNeedsFormfeed(x.X) || f.nextNeedsFormfeed(x.Y)
+	case *ast.IndexExpr:
+		return f.nextNeedsFormfeed(x.X)
+	case *ast.SelectorExpr:
+		return f.nextNeedsFormfeed(x.X)
 	case *ast.CallExpr:
 		for _, arg := range x.Args {
 			if f.nextNeedsFormfeed(arg) {
@@ -469,7 +506,7 @@ func (f *formatter) label(l ast.Label, constraint token.Token) {
 		// if the AST is not generated by the parser.
 		name := n.Name
 		if !ast.IsValidIdent(name) {
-			name = literal.String.Quote(n.Name)
+			name = literal.Label.Quote(n.Name)
 		}
 		f.print(n.NamePos, name)
 
@@ -479,7 +516,7 @@ func (f *formatter) label(l ast.Label, constraint token.Token) {
 		// according to spec.
 		if strings.HasPrefix(str, `"""`) || strings.HasPrefix(str, "#") {
 			if u, err := literal.Unquote(str); err == nil {
-				str = literal.String.Quote(u)
+				str = literal.Label.Quote(u)
 			}
 		}
 		f.print(n.ValuePos, str)
@@ -631,20 +668,20 @@ func (f *formatter) exprRaw(expr ast.Expr, prec1, depth int) {
 
 		switch {
 		case len(x.Elts) == 0:
-			if !x.Rbrace.HasRelPos() {
-				// collapse curly braces if the body is empty.
-				ffAlt := blank | nooverride
-				for _, c := range x.Comments() {
-					if c.Position == 1 {
-						ffAlt = ff
-					}
+			// collapse curly braces if the body is empty.
+			ffAlt := blank | nooverride
+			for _, c := range x.Comments() {
+				if c.Position == 1 {
+					ffAlt = ff
+					break
 				}
-				ff = ffAlt
 			}
+			ff = ffAlt
 		case !x.Rbrace.HasRelPos() || !x.Elts[0].Pos().HasRelPos():
 			ws |= newline | nooverride
 		}
 		f.print(x.Lbrace, token.LBRACE, &l, ws, ff, indent)
+		f.prevLbraceOnLine = l == f.lineout
 
 		f.walkDeclList(x.Elts)
 		f.matchUnindent()
@@ -659,7 +696,20 @@ func (f *formatter) exprRaw(expr ast.Expr, prec1, depth int) {
 		f.print(ws, x.Rbrace, token.RBRACE)
 
 	case *ast.ListLit:
-		f.print(x.Lbrack, token.LBRACK, noblank, indent)
+		ws := noblank | indent
+		if len(x.Elts) == 0 {
+			// collapse square brackets if the body is empty.
+			collapseWs := blank | nooverride
+			for _, c := range x.Comments() {
+				if c.Position == 1 {
+					collapseWs = ws
+					break
+				}
+			}
+			ws |= collapseWs
+		}
+
+		f.print(x.Lbrack, token.LBRACK, ws)
 		f.walkListElems(x.Elts)
 		f.print(trailcomma, noblank)
 		f.visitComments(f.current.pos)
@@ -698,6 +748,8 @@ func (f *formatter) clause(clause ast.Clause) {
 		f.markUnindentLine()
 
 	case *ast.LetClause:
+		// TODO(mvdan): LetClause is handled in both the clause and decl methods,
+		// because at the semantic level it is different in each case, but the code is repetitive.
 		f.print(n.Let, token.LET, blank, nooverride)
 		f.print(indent)
 		f.expr(n.Ident)
@@ -849,7 +901,7 @@ func (f *formatter) binaryExpr(x *ast.BinaryExpr, prec1, cutoff, depth int) {
 	prec := x.Op.Precedence()
 	if prec < prec1 {
 		// parenthesis needed
-		// Note: The parser inserts an syntax.ParenExpr node; thus this case
+		// Note: The parser inserts a syntax.ParenExpr node; thus this case
 		//       can only occur if the AST is created in a different way.
 		// defer p.pushComment(nil).pop()
 		f.print(token.LPAREN, nooverride)
