@@ -17,13 +17,16 @@ package load
 import (
 	"context"
 	"fmt"
+	"io/fs"
 	"sort"
 	"strconv"
+	"strings"
 
 	"cuelang.org/go/cue/ast"
 	"cuelang.org/go/cue/build"
 	"cuelang.org/go/internal/cueexperiment"
 	"cuelang.org/go/internal/filetypes"
+	"cuelang.org/go/internal/mod/modimports"
 	"cuelang.org/go/internal/mod/modpkgload"
 	"cuelang.org/go/internal/mod/modrequirements"
 	"cuelang.org/go/mod/module"
@@ -58,7 +61,6 @@ func Instances(args []string, c *Config) []*build.Instance {
 	if len(args) == 0 {
 		args = []string{"."}
 	}
-
 	// TODO: This requires packages to be placed before files. At some point this
 	// could be relaxed.
 	i := 0
@@ -75,7 +77,25 @@ func Instances(args []string, c *Config) []*build.Instance {
 			return []*build.Instance{c.newErrInstance(err)}
 		}
 	}
-	synCache := newSyntaxCache(c)
+	if c.Package != "" && c.Package != "_" && c.Package != "*" {
+		// The caller has specified an explicit package to load.
+		// This is essentially the same as passing an explicit package
+		// qualifier to all package arguments that don't already have
+		// one. We add that qualifier here so that there's a distinction
+		// between package paths specified as arguments, which
+		// have the qualifier added, and package paths that are dependencies
+		// of those, which don't.
+		pkgArgs1 := make([]string, 0, len(pkgArgs))
+		for _, p := range pkgArgs {
+			if ip := module.ParseImportPath(p); !ip.ExplicitQualifier {
+				ip.Qualifier = c.Package
+				p = ip.String()
+			}
+			pkgArgs1 = append(pkgArgs1, p)
+		}
+		pkgArgs = pkgArgs1
+	}
+
 	tg := newTagger(c)
 	// Pass all arguments that look like packages to loadPackages
 	// so that they'll be available when looking up the packages
@@ -84,17 +104,24 @@ func Instances(args []string, c *Config) []*build.Instance {
 	if err != nil {
 		return []*build.Instance{c.newErrInstance(err)}
 	}
-	pkgs, err := loadPackages(ctx, c, synCache, expandedPaths, otherFiles)
-	if err != nil {
-		return []*build.Instance{c.newErrInstance(err)}
+
+	var pkgs *modpkgload.Packages
+	if !c.SkipImports {
+		pkgs, err = loadPackages(ctx, c, expandedPaths, otherFiles, tg)
+		if err != nil {
+			return []*build.Instance{c.newErrInstance(err)}
+		}
 	}
-	l := newLoader(c, tg, synCache, pkgs)
+	l := newLoader(c, tg, pkgs)
 
 	if c.Context == nil {
-		c.Context = build.NewContext(
-			build.Loader(l.loadFunc),
+		opts := []build.Option{
 			build.ParseFile(c.ParseFile),
-		)
+		}
+		if f := l.loadFunc(); l != nil {
+			opts = append(opts, build.Loader(f))
+		}
+		c.Context = build.NewContext(opts...)
 	}
 
 	a := []*build.Instance{}
@@ -151,12 +178,19 @@ func Instances(args []string, c *Config) []*build.Instance {
 
 // loadPackages returns packages loaded from the given package list and also
 // including imports from the given build files.
-func loadPackages(ctx context.Context, cfg *Config, synCache *syntaxCache, pkgs []resolvedPackageArg, otherFiles []*build.File) (*modpkgload.Packages, error) {
+func loadPackages(
+	ctx context.Context,
+	cfg *Config,
+	pkgs []resolvedPackageArg,
+	otherFiles []*build.File,
+	tg *tagger,
+) (*modpkgload.Packages, error) {
 	if cfg.Registry == nil || cfg.modFile == nil || cfg.modFile.Module == "" {
 		return nil, nil
 	}
+	mainModPath := cfg.modFile.QualifiedModule()
 	reqs := modrequirements.NewRequirements(
-		cfg.modFile.Module,
+		mainModPath,
 		cfg.Registry,
 		cfg.modFile.DepVersions(),
 		cfg.modFile.DefaultMajorVersions(),
@@ -176,21 +210,19 @@ func loadPackages(ctx context.Context, cfg *Config, synCache *syntaxCache, pkgs
 			// not a CUE file; assume it has no imports for now.
 			continue
 		}
-		syntaxes, err := synCache.getSyntax(f)
+		syntax, err := cfg.fileSystem.getCUESyntax(f)
 		if err != nil {
-			return nil, fmt.Errorf("cannot get syntax for %q: %v", f.Filename, err)
+			return nil, fmt.Errorf("cannot get syntax for %q: %w", f.Filename, err)
 		}
-		for _, syntax := range syntaxes {
-			for _, imp := range syntax.Imports {
-				pkgPath, err := strconv.Unquote(imp.Path.Value)
-				if err != nil {
-					// Should never happen.
-					return nil, fmt.Errorf("invalid import path %q in %s", imp.Path.Value, f.Filename)
-				}
-				// Canonicalize the path.
-				pkgPath = module.ParseImportPath(pkgPath).Canonical().String()
-				pkgPaths[pkgPath] = true
+		for _, imp := range syntax.Imports {
+			pkgPath, err := strconv.Unquote(imp.Path.Value)
+			if err != nil {
+				// Should never happen.
+				return nil, fmt.Errorf("invalid import path %q in %s", imp.Path.Value, f.Filename)
 			}
+			// Canonicalize the path.
+			pkgPath = module.ParseImportPath(pkgPath).Canonical().String()
+			pkgPaths[pkgPath] = true
 		}
 	}
 	// TODO use maps.Keys when we can.
@@ -206,5 +238,34 @@ func loadPackages(ctx context.Context, cfg *Config, synCache *syntaxCache, pkgs
 		reqs,
 		cfg.Registry,
 		pkgPathSlice,
+		func(pkgPath string, mod module.Version, fsys fs.FS, mf modimports.ModuleFile) bool {
+			if !cfg.Tools && strings.HasSuffix(mf.FilePath, "_tool.cue") {
+				return false
+			}
+			isTest := strings.HasSuffix(mf.FilePath, "_test.cue")
+			var tagIsSet func(string) bool
+			if mod.Path() == mainModPath {
+				// In the main module.
+				if isTest && !cfg.Tests {
+					return false
+				}
+				tagIsSet = tg.tagIsSet
+			} else {
+				// Outside the main module.
+				if isTest {
+					// Don't traverse test files outside the main module
+					return false
+				}
+				// Treat all build tag keys as unset.
+				tagIsSet = func(string) bool {
+					return false
+				}
+			}
+			if err := shouldBuildFile(mf.Syntax, tagIsSet); err != nil {
+				// Later build logic should pick up and report the same error.
+				return false
+			}
+			return true
+		},
 	), nil
 }
