@@ -79,7 +79,7 @@ const (
 	// Kind.String to indicate NumberKind.
 
 	// NumberKind represents any kind of number.
-	NumberKind Kind = IntKind | FloatKind
+	NumberKind Kind = adt.NumberKind
 
 	// TopKind represents the top value.
 	TopKind Kind = adt.TopKind
@@ -146,7 +146,7 @@ func (o *hiddenStructValue) Lookup(key string) Value {
 func (o *structValue) marshalJSON() (b []byte, err errors.Error) {
 	b = append(b, '{')
 	n := o.Len()
-	for i := 0; i < n; i++ {
+	for i := range n {
 		k, v := o.At(i)
 		s, err := internaljson.Marshal(k)
 		if err != nil {
@@ -223,8 +223,8 @@ type Iterator struct {
 
 type hiddenIterator = Iterator
 
-// Next advances the iterator to the next value and reports whether there was
-// any. It must be called before the first call to Value or Key.
+// Next advances the iterator to the next value and reports whether there was any.
+// It must be called before the first call to [Iterator.Value] or [Iterator.Selector].
 func (i *Iterator) Next() bool {
 	if i.p >= len(i.arcs) {
 		i.cur = Value{}
@@ -233,15 +233,15 @@ func (i *Iterator) Next() bool {
 	arc := i.arcs[i.p]
 	arc.Finalize(i.ctx)
 	p := linkParent(i.val.parent_, i.val.v, arc)
-	i.cur = makeValue(i.val.idx, arc, p)
 	i.f = arc.Label
 	i.arcType = arc.ArcType
+	i.cur = makeValue(i.val.idx, arc, p)
 	i.p++
 	return true
 }
 
-// Value returns the current value in the list. It will panic if Next advanced
-// past the last entry.
+// Value returns the current value in the list.
+// It will panic if [Iterator.Next] advanced past the last entry.
 func (i *Iterator) Value() Value {
 	return i.cur
 }
@@ -259,8 +259,8 @@ func (i *Iterator) Selector() Selector {
 // Label reports the label of the value if i iterates over struct fields and ""
 // otherwise.
 //
-// Slated to be deprecated: use i.Selector().String(). Note that this will give
-// more accurate string representations.
+// Slated to be deprecated: use [Iterator.Selector] and [Selector.String].
+// Note that this will give more accurate string representations.
 func (i *hiddenIterator) Label() string {
 	if i.f == 0 {
 		return ""
@@ -332,7 +332,7 @@ func (v Value) getNum(k adt.Kind) (*adt.Num, errors.Error) {
 // 200 and exp == -2. Calling MantExp with a nil argument is an efficient way to
 // get the exponent of the receiver.
 func (v Value) MantExp(mant *big.Int) (exp int, err error) {
-	n, err := v.getNum(adt.NumKind)
+	n, err := v.getNum(adt.NumberKind)
 	if err != nil {
 		return 0, err
 	}
@@ -351,7 +351,7 @@ func (v Value) MantExp(mant *big.Int) (exp int, err error) {
 // Decimal is for internal use only. The Decimal type that is returned is
 // subject to change.
 func (v hiddenValue) Decimal() (d *internal.Decimal, err error) {
-	n, err := v.getNum(adt.NumKind)
+	n, err := v.getNum(adt.NumberKind)
 	if err != nil {
 		return nil, err
 	}
@@ -372,7 +372,7 @@ func (v Value) AppendInt(buf []byte, base int) ([]byte, error) {
 // AppendFloat appends to buf the string form of the floating-point number x.
 // It returns an error if v is not a number.
 func (v Value) AppendFloat(buf []byte, fmt byte, prec int) ([]byte, error) {
-	n, err := v.getNum(adt.NumKind)
+	n, err := v.getNum(adt.NumberKind)
 	if err != nil {
 		return nil, err
 	}
@@ -468,23 +468,6 @@ func (v Value) Uint64() (uint64, error) {
 	return i, nil
 }
 
-// trimZeros trims 0's for better JSON representations.
-func trimZeros(s string) string {
-	n1 := len(s)
-	s2 := strings.TrimRight(s, "0")
-	n2 := len(s2)
-	if p := strings.IndexByte(s2, '.'); p != -1 {
-		if p == n2-1 {
-			return s[:len(s2)+1]
-		}
-		return s2
-	}
-	if n1-n2 <= 4 {
-		return s
-	}
-	return fmt.Sprint(s2, "e+", n1-n2)
-}
-
 var (
 	smallestPosFloat64 *apd.Decimal
 	smallestNegFloat64 *apd.Decimal
@@ -527,7 +510,7 @@ func init() {
 // by a float64 (|x| > math.MaxFloat64), the result is (+Inf, ErrAbove) or
 // (-Inf, ErrBelow), depending on the sign of x.
 func (v Value) Float64() (float64, error) {
-	n, err := v.getNum(adt.NumKind)
+	n, err := v.getNum(adt.NumberKind)
 	if err != nil {
 		return 0, err
 	}
@@ -634,6 +617,7 @@ func newValueRoot(idx *runtime.Runtime, ctx *adt.OpContext, x adt.Expr) Value {
 
 func newChildValue(o *structValue, i int) Value {
 	arc := o.at(i)
+	// TODO: fix linkage to parent.
 	return makeValue(o.v.idx, arc, linkParent(o.v.parent_, o.v.v, arc))
 }
 
@@ -641,11 +625,16 @@ func newChildValue(o *structValue, i int) Value {
 // otherwise.
 func Dereference(v Value) Value {
 	n := v.v
-	if n == nil || len(n.Conjuncts) != 1 {
+	if n == nil {
 		return v
 	}
 
-	env, expr := n.Conjuncts[0].EnvExpr()
+	c, count := n.SingleConjunct()
+	if count != 1 {
+		return v
+	}
+
+	env, expr := c.EnvExpr()
 
 	// TODO: consider supporting unwrapping of structs or comprehensions around
 	// a single embedded reference.
@@ -654,7 +643,7 @@ func Dereference(v Value) Value {
 		return v
 	}
 
-	c := adt.MakeRootConjunct(env, expr)
+	c = adt.MakeRootConjunct(env, expr)
 
 	ctx := v.ctx()
 	n, b := ctx.Resolve(c, r)
@@ -940,7 +929,7 @@ func (v Value) marshalJSON() (b []byte, err error) {
 		return internaljson.Marshal(nil)
 	case adt.BoolKind:
 		return internaljson.Marshal(x.(*adt.Bool).B)
-	case adt.IntKind, adt.FloatKind, adt.NumKind:
+	case adt.IntKind, adt.FloatKind, adt.NumberKind:
 		b, err := x.(*adt.Num).X.MarshalText()
 		b = bytes.TrimLeft(b, "+")
 		return b, err
@@ -986,6 +975,7 @@ func (v Value) Syntax(opts ...Option) ast.Node {
 		ShowDocs:        o.docs,
 		ShowErrors:      o.showErrors,
 		InlineImports:   o.inlineImports,
+		Fragment:        o.raw,
 	}
 
 	pkgID := v.instance().ID()
@@ -1071,10 +1061,11 @@ func (v hiddenValue) Split() []Value {
 		return nil
 	}
 	a := []Value{}
-	for _, x := range v.v.Conjuncts {
+	v.v.VisitLeafConjuncts(func(x adt.Conjunct) bool {
 		env, expr := x.EnvExpr()
 		a = append(a, remakeValue(v, env, expr))
-	}
+		return true
+	})
 	return a
 }
 
@@ -1086,10 +1077,17 @@ func (v Value) Source() ast.Node {
 	if v.v == nil {
 		return nil
 	}
-	if len(v.v.Conjuncts) == 1 {
-		return v.v.Conjuncts[0].Source()
+	count := 0
+	var src ast.Node
+	v.v.VisitLeafConjuncts(func(c adt.Conjunct) bool {
+		src = c.Source()
+		count++
+		return true
+	})
+	if count > 1 || src == nil {
+		src = v.v.Value().Source()
 	}
-	return v.v.Value().Source()
+	return src
 }
 
 // If v exactly represents a package, BuildInstance returns
@@ -1127,24 +1125,25 @@ func (v Value) Pos() token.Pos {
 	}
 	// Pick the most-concrete field.
 	var p token.Pos
-	for _, c := range v.v.Conjuncts {
+	v.v.VisitLeafConjuncts(func(c adt.Conjunct) bool {
 		x := c.Elem()
 		pp := pos(x)
 		if pp == token.NoPos {
-			continue
+			return true
 		}
 		p = pp
 		// Prefer struct conjuncts with actual fields.
 		if s, ok := x.(*adt.StructLit); ok && len(s.Fields) > 0 {
-			break
+			return false
 		}
-	}
+		return true
+	})
 	return p
 }
 
 // TODO: IsFinal: this value can never be changed.
 
-// IsClosed reports whether a list of struct is closed. It reports false when
+// IsClosed reports whether a list or struct is closed. It reports false when
 // the value is not a list or struct.
 //
 // Deprecated: use Allows(AnyString) and Allows(AnyIndex) or Kind/IncompleteKind.
@@ -1156,6 +1155,8 @@ func (v hiddenValue) IsClosed() bool {
 	case ListKind:
 		return v.v.IsClosedList()
 	case StructKind:
+		// TODO: remove this more expensive computation once the old evaluator
+		// is removed.
 		return !v.Allows(AnyString)
 	}
 	return false
@@ -1166,6 +1167,9 @@ func (v hiddenValue) IsClosed() bool {
 // Allows does not take into account validators like list.MaxItems(4). This may
 // change in the future.
 func (v Value) Allows(sel Selector) bool {
+	if v.v.HasEllipsis {
+		return true
+	}
 	c := v.ctx()
 	f := sel.sel.feature(c)
 	return v.v.Accept(c, f)
@@ -1179,7 +1183,7 @@ func (v Value) IsConcrete() bool {
 	if v.v == nil {
 		return false // any is neither concrete, not a list or struct.
 	}
-	if b, ok := v.v.BaseValue.(*adt.Bottom); ok {
+	if b := v.v.Bottom(); b != nil {
 		return !b.IsIncomplete()
 	}
 	if !adt.IsConcrete(v.v) {
@@ -1204,12 +1208,37 @@ func (v Value) Exists() bool {
 	if v.v == nil {
 		return false
 	}
-	if err, ok := v.v.BaseValue.(*adt.Bottom); ok {
+	if err := v.v.Bottom(); err != nil {
 		return !err.NotExists
 	}
 	return true
 }
 
+// isKind reports whether a value matches a particular kind.
+// It is like checkKind, except that it doesn't construct an error value.
+// Note that when v is bottom, the method always returns false.
+func (v Value) isKind(ctx *adt.OpContext, want adt.Kind) bool {
+	if v.v == nil {
+		return false
+	}
+	x := v.eval(ctx)
+	if _, ok := x.(*adt.Bottom); ok {
+		return false
+	}
+	k := x.Kind()
+	if want != adt.BottomKind {
+		if k&want == adt.BottomKind {
+			return false
+		}
+		if !adt.IsConcrete(x) {
+			return false
+		}
+	}
+	return true
+}
+
+// checkKind returns a bottom error if a value does not match a particular kind,
+// describing the reason why. Note that when v is bottom, it is always returned as-is.
 func (v Value) checkKind(ctx *adt.OpContext, want adt.Kind) *adt.Bottom {
 	if v.v == nil {
 		return errNotExists
@@ -1275,7 +1304,7 @@ func (v Value) Len() Value {
 
 // Elem returns the value of undefined element types of lists and structs.
 //
-// Deprecated: use LookupPath in combination with "AnyString" or "AnyIndex".
+// Deprecated: use [Value.LookupPath] in combination with "AnyString" or "AnyIndex".
 func (v hiddenValue) Elem() (Value, bool) {
 	sel := AnyString
 	if v.v.IsList() {
@@ -1311,10 +1340,11 @@ func (v Value) Null() error {
 	return nil
 }
 
-// // IsNull reports whether v is null.
-// func (v Value) IsNull() bool {
-// 	return v.Null() == nil
-// }
+// IsNull reports whether v is null.
+func (v Value) IsNull() bool {
+	v, _ = v.Default()
+	return v.isKind(v.ctx(), adt.NullKind)
+}
 
 // Bool returns the bool value of v or false and an error if v is not a boolean.
 func (v Value) Bool() (bool, error) {
@@ -1383,8 +1413,8 @@ func (v Value) structValOpts(ctx *adt.OpContext, o options) (s structValue, err
 
 	obj := v.v
 
-	switch b, ok := v.v.BaseValue.(*adt.Bottom); {
-	case ok && b.IsIncomplete() && !o.concrete && !o.final:
+	switch b := v.v.Bottom(); {
+	case b != nil && b.IsIncomplete() && !o.concrete && !o.final:
 
 	// Allow scalar values if hidden or definition fields are requested.
 	case !o.omitHidden, !o.omitDefinitions:
@@ -1411,7 +1441,7 @@ func (v Value) structValOpts(ctx *adt.OpContext, o options) (s structValue, err
 		if f.IsHidden() && o.omitHidden {
 			continue
 		}
-		arc := obj.Lookup(f)
+		arc := obj.LookupRaw(f)
 		if arc == nil {
 			continue
 		}
@@ -1426,7 +1456,7 @@ func (v Value) structValOpts(ctx *adt.OpContext, o options) (s structValue, err
 			// it avoids hiding errors in required fields.
 			if o.omitOptional || o.concrete || o.final {
 				arc = &adt.Vertex{
-					Label:     arc.Label,
+					Label:     f,
 					Parent:    arc.Parent,
 					Conjuncts: arc.Conjuncts,
 					BaseValue: adt.NewRequiredNotPresentError(ctx, arc),
@@ -1442,7 +1472,7 @@ func (v Value) structValOpts(ctx *adt.OpContext, o options) (s structValue, err
 // Struct returns the underlying struct of a value or an error if the value
 // is not a struct.
 //
-// Deprecated: use Fields.
+// Deprecated: use [Value.Fields].
 func (v hiddenValue) Struct() (*Struct, error) {
 	ctx := v.ctx()
 	obj, err := v.structValOpts(ctx, options{})
@@ -1611,7 +1641,7 @@ func appendPath(a []Selector, v Value) []Selector {
 
 // LookupDef is equal to LookupPath(MakePath(Def(name))).
 //
-// Deprecated: use LookupPath.
+// Deprecated: use [Value.LookupPath].
 func (v hiddenValue) LookupDef(name string) Value {
 	return v.LookupPath(MakePath(Def(name)))
 }
@@ -1622,7 +1652,7 @@ var errNotFound = errors.Newf(token.NoPos, "field not found")
 // look up a definition or hidden field (starting with `_` or `_#`). Otherwise
 // it interprets name as an arbitrary string for a regular field.
 //
-// Deprecated: use LookupPath.
+// Deprecated: use [Value.LookupPath].
 func (v hiddenValue) FieldByName(name string, isIdent bool) (f FieldInfo, err error) {
 	s, err := v.Struct()
 	if err != nil {
@@ -1633,7 +1663,7 @@ func (v hiddenValue) FieldByName(name string, isIdent bool) (f FieldInfo, err er
 
 // LookupField reports information about a field of v.
 //
-// Deprecated: use LookupPath
+// Deprecated: use [Value.LookupPath].
 func (v hiddenValue) LookupField(name string) (FieldInfo, error) {
 	s, err := v.Struct()
 	if err != nil {
@@ -1671,7 +1701,7 @@ func (v hiddenValue) LookupField(name string) (FieldInfo, error) {
 // Any reference in v referring to the value at the given path will resolve
 // to x in the newly created value. The resulting value is not validated.
 //
-// Deprecated: use FillPath.
+// Deprecated: use [Value.FillPath].
 func (v hiddenValue) Fill(x interface{}, path ...string) Value {
 	if v.v == nil {
 		return v
@@ -1741,7 +1771,7 @@ func (v Value) FillPath(p Path, x interface{}) Value {
 			list := &adt.ListLit{}
 			any := &adt.Top{}
 			// TODO(perf): make this a constant thing. This will be possible with the query extension.
-			for k := 0; k < i; k++ {
+			for range i {
 				list.Elems = append(list.Elems, any)
 			}
 			list.Elems = append(list.Elems, expr, &adt.Ellipsis{})
@@ -1777,14 +1807,18 @@ func (v Value) FillPath(p Path, x interface{}) Value {
 // The returned function returns the value that would be unified with field
 // given its name.
 //
-// Deprecated: use LookupPath in combination with using optional selectors.
+// Deprecated: use [Value.LookupPath] in combination with using optional selectors.
 func (v hiddenValue) Template() func(label string) Value {
 	if v.v == nil {
 		return nil
 	}
 
+	// Implementation for the old evaluator.
 	types := v.v.OptionalTypes()
-	if types&(adt.HasAdditional|adt.HasPattern) == 0 {
+	switch {
+	case types&(adt.HasAdditional|adt.HasPattern) != 0:
+	case v.v.PatternConstraints != nil:
+	default:
 		return nil
 	}
 
@@ -1983,11 +2017,11 @@ func (v hiddenValue) Reference() (inst *Instance, path []string) {
 // is not a reference.
 func (v Value) ReferencePath() (root Value, p Path) {
 	// TODO: don't include references to hidden fields.
-	if v.v == nil || len(v.v.Conjuncts) != 1 {
+	c, count := v.v.SingleConjunct()
+	if count != 1 {
 		return Value{}, Path{}
 	}
 	ctx := v.ctx()
-	c := v.v.Conjuncts[0]
 
 	env, expr := c.EnvExpr()
 
@@ -2128,7 +2162,7 @@ func DisallowCycles(disallow bool) Option {
 
 // ResolveReferences forces the evaluation of references when outputting.
 //
-// Deprecated: Syntax will now always attempt to resolve dangling references and
+// Deprecated: [Value.Syntax] will now always attempt to resolve dangling references and
 // make the output self-contained. When [Final] or [Concrete] are used,
 // it will already attempt to resolve all references.
 // See also [InlineImports].
@@ -2157,7 +2191,13 @@ func ErrorsAsValues(show bool) Option {
 	return func(p *options) { p.showErrors = show }
 }
 
-// Raw tells Syntax to generate the value as is without any simplifications.
+// Raw tells Syntax to generate the value as is without any simplifications and
+// without ensuring a value is self contained. Any references are left dangling.
+// The generated syntax tree can be compiled by passing the Value from which it
+// was generated to scope.
+//
+// The option InlineImports overrides this option with respect to ensuring the
+// output is self contained.
 func Raw() Option {
 	return func(p *options) { p.raw = true }
 }
@@ -2259,7 +2299,7 @@ func (v Value) Walk(before func(Value) bool, after func(Value)) {
 			omitHidden:      true,
 			omitDefinitions: true,
 		})
-		for i := 0; i < obj.Len(); i++ {
+		for i := range obj.Len() {
 			_, v := obj.At(i)
 			// TODO: should we error on required fields, or visit them anyway?
 			// Walk is not designed to error at this moment, though.
@@ -2315,41 +2355,45 @@ func (v Value) Expr() (Op, []Value) {
 
 	if v.v.IsData() {
 		expr = v.v.Value()
+		goto process
 
-	} else {
-		switch len(v.v.Conjuncts) {
-		case 0:
-			if v.v.BaseValue == nil {
-				return NoOp, []Value{makeValue(v.idx, v.v, v.parent_)} // TODO: v?
-			}
-			expr = v.v.Value()
-
-		case 1:
-			// the default case, processed below.
-			c := v.v.Conjuncts[0]
-			env, expr = c.EnvExpr()
-			if w, ok := expr.(*adt.Vertex); ok {
-				return Value{v.idx, w, v.parent_}.Expr()
-			}
-
-		default:
-			a := []Value{}
-			ctx := v.ctx()
-			for _, c := range v.v.Conjuncts {
-				// Keep parent here. TODO: do we need remove the requirement
-				// from other conjuncts?
-				n := &adt.Vertex{
-					Parent: v.v.Parent,
-					Label:  v.v.Label,
-				}
-				n.AddConjunct(c)
-				n.Finalize(ctx)
-				a = append(a, makeValue(v.idx, n, v.parent_))
-			}
-			return adt.AndOp, a
-		}
 	}
 
+	switch c, count := v.v.SingleConjunct(); count {
+	case 0:
+		if v.v.BaseValue == nil {
+			return NoOp, []Value{makeValue(v.idx, v.v, v.parent_)} // TODO: v?
+		}
+		expr = v.v.Value()
+
+	case 1:
+		// the default case, processed below.
+		env, expr = c.EnvExpr()
+		if w, ok := expr.(*adt.Vertex); ok {
+			return Value{v.idx, w, v.parent_}.Expr()
+		}
+
+	default:
+		a := []Value{}
+		ctx := v.ctx()
+		v.v.VisitLeafConjuncts(func(c adt.Conjunct) bool {
+			// Keep parent here. TODO: do we need remove the requirement
+			// from other conjuncts?
+			n := &adt.Vertex{
+				Parent: v.v.Parent,
+				Label:  v.v.Label,
+			}
+			n.AddConjunct(c)
+			n.Finalize(ctx)
+			a = append(a, makeValue(v.idx, n, v.parent_))
+			return true
+		})
+
+		return adt.AndOp, a
+	}
+
+process:
+
 	// TODO: replace appends with []Value{}. For not leave.
 	a := []Value{}
 	op := NoOp
