@@ -17,7 +17,6 @@ package adt
 import (
 	"fmt"
 	"log"
-	"os"
 	"reflect"
 	"regexp"
 	"sort"
@@ -31,25 +30,17 @@ import (
 	"cuelang.org/go/cue/stats"
 	"cuelang.org/go/cue/token"
 	"cuelang.org/go/internal"
+	"cuelang.org/go/internal/cuedebug"
 )
 
-// Debug sets whether extra aggressive checking should be done.
-// This should typically default to true for pre-releases and default to
-// false otherwise.
-var Debug bool = os.Getenv("CUE_DEBUG") != "0"
-
-// Verbosity sets the log level. There are currently only two levels:
-//
-//	0: no logging
-//	1: logging
-var Verbosity int
-
 // DebugSort specifies that arcs be sorted consistently between implementations.
 //
 //	0: default
 //	1: sort by Feature: this should be consistent between implementations where
 //		   there is no change in the compiler and indexing code.
 //	2: alphabetical
+//
+// TODO: move to DebugFlags
 var DebugSort int
 
 func DebugSortArcs(c *OpContext, n *Vertex) {
@@ -92,8 +83,8 @@ func DebugSortFields(c *OpContext, a []Feature) {
 //
 // It is advisable for each use of Assert to document how the error is expected
 // to be handled down the line.
-func Assertf(b bool, format string, args ...interface{}) {
-	if Debug && !b {
+func Assertf(c *OpContext, b bool, format string, args ...interface{}) {
+	if c.Strict && !b {
 		panic(fmt.Sprintf("assertion failed: "+format, args...))
 	}
 }
@@ -101,7 +92,7 @@ func Assertf(b bool, format string, args ...interface{}) {
 // Assertf either panics or reports an error to c if the condition is not met.
 func (c *OpContext) Assertf(pos token.Pos, b bool, format string, args ...interface{}) {
 	if !b {
-		if Debug {
+		if c.Strict {
 			panic(fmt.Sprintf("assertion failed: "+format, args...))
 		}
 		c.addErrf(0, pos, format, args...)
@@ -115,7 +106,7 @@ func init() {
 var pMap = map[*Vertex]int{}
 
 func (c *OpContext) Logf(v *Vertex, format string, args ...interface{}) {
-	if Verbosity == 0 {
+	if c.LogEval == 0 {
 		return
 	}
 	if v == nil {
@@ -176,12 +167,12 @@ type Runtime interface {
 	// type if available.
 	LoadType(t reflect.Type) (src ast.Expr, expr Expr, ok bool)
 
-	EvaluatorVersion() internal.EvaluatorVersion
+	Settings() (internal.EvaluatorVersion, cuedebug.Config)
 }
 
 type Config struct {
 	Runtime
-	Format func(Node) string
+	Format func(Runtime, Node) string
 }
 
 // New creates an operation context.
@@ -189,11 +180,13 @@ func New(v *Vertex, cfg *Config) *OpContext {
 	if cfg.Runtime == nil {
 		panic("nil Runtime")
 	}
+	version, flags := cfg.Runtime.Settings()
 	ctx := &OpContext{
 		Runtime:     cfg.Runtime,
 		Format:      cfg.Format,
 		vertex:      v,
-		Version:     cfg.Runtime.EvaluatorVersion(),
+		Version:     version,
+		Config:      flags,
 		taskContext: schedConfig,
 	}
 	if v != nil {
@@ -212,8 +205,9 @@ func (c *OpContext) isDevVersion() bool {
 // use an OpContext at a time.
 type OpContext struct {
 	Runtime
-	Format func(Node) string
+	Format func(Runtime, Node) string
 
+	cuedebug.Config
 	Version internal.EvaluatorVersion // Copied from Runtime
 
 	taskContext
@@ -246,6 +240,15 @@ type OpContext struct {
 	closed     map[*closeInfo]*closeStats
 	todo       *closeStats
 
+	// evalDepth indicates the current depth of evaluation. It is used to
+	// detect structural cycles and their severity.s
+	evalDepth int
+
+	// optionalMark indicates the evalDepth at which the last optional field,
+	// pattern constraint or other construct that may contain errors was
+	// encountered. A value of 0 indicates we are not within such field.
+	optionalMark int
+
 	// inDisjunct indicates that non-monotonic checks should be skipped.
 	// This is used if we want to do some extra work to eliminate disjunctions
 	// early. The result of unification should be thrown away if this check is
@@ -493,7 +496,9 @@ func (c *OpContext) resolveState(x Conjunct, r Resolver, state combinedFlags) (*
 		return nil, arc.ChildErrors
 	}
 
-	arc = arc.Indirect()
+	// Dereference any vertices that do not contribute to more knownledge about
+	// the node.
+	arc = arc.DerefNonRooted()
 
 	return arc, err
 }
@@ -506,8 +511,10 @@ func (c *OpContext) Lookup(env *Environment, r Resolver) (*Vertex, *Bottom) {
 
 	err := c.PopState(s)
 
-	if arc != nil {
-		arc = arc.Indirect()
+	if arc != nil && !c.isDevVersion() {
+		// TODO(deref): lookup should probably not use DerefValue, but
+		// rather only dereference disjunctions.
+		arc = arc.DerefValue()
 	}
 
 	return arc, err
@@ -628,7 +635,7 @@ func (c *OpContext) evaluateRec(v Conjunct, state combinedFlags) Value {
 	val := c.evalState(x, state)
 	if val == nil {
 		// Be defensive: this never happens, but just in case.
-		Assertf(false, "nil return value: unspecified error")
+		Assertf(c, false, "nil return value: unspecified error")
 		val = &Bottom{
 			Code: IncompleteError,
 			Err:  c.Newf("UNANTICIPATED ERROR"),
@@ -660,7 +667,7 @@ func (c *OpContext) evalState(v Expr, state combinedFlags) (result Value) {
 		c.errs = CombineErrors(c.src, c.errs, err)
 
 		if v, ok := result.(*Vertex); ok {
-			if b, _ := v.BaseValue.(*Bottom); b != nil {
+			if b := v.Bottom(); b != nil {
 				switch b.Code {
 				case IncompleteError:
 				case CycleError:
@@ -703,11 +710,23 @@ func (c *OpContext) evalState(v Expr, state combinedFlags) (result Value) {
 		if arc == nil {
 			return nil
 		}
+		// TODO(deref): what is the right level of dereferencing here?
+		// DerefValue seems to work too.
+		arc = arc.DerefNonShared()
+
+		// TODO: consider moving this after markCycle, depending on how we
+		// implement markCycle, or whether we need it at all.
+		// TODO: is this indirect necessary?
+		// arc = arc.Indirect()
 
 		// Save the old CloseInfo and restore after evaluate to avoid detecting
 		// spurious cycles.
 		saved := c.ci
-		if n := arc.state; n != nil {
+		n := arc.state
+		if c.isDevVersion() {
+			n = arc.getState(c)
+		}
+		if n != nil {
 			c.ci, _ = n.markCycle(arc, nil, x, c.ci)
 		}
 		c.ci.Inline = true
@@ -784,7 +803,7 @@ func (c *OpContext) unifyNode(v Expr, state combinedFlags) (result Value) {
 		c.errs = CombineErrors(c.src, c.errs, err)
 
 		if v, ok := result.(*Vertex); ok {
-			if b, _ := v.BaseValue.(*Bottom); b != nil && !b.IsIncomplete() {
+			if b := v.Bottom(); b != nil && !b.IsIncomplete() {
 				result = b
 			}
 		}
@@ -823,6 +842,12 @@ func (c *OpContext) unifyNode(v Expr, state combinedFlags) (result Value) {
 		if v == nil {
 			return nil
 		}
+		v = v.DerefValue()
+
+		// TODO: consider moving this after markCycle, depending on how we
+		// implement markCycle, or whether we need it at all.
+		// TODO: is this indirect necessary?
+		// v = v.Indirect()
 
 		if c.isDevVersion() {
 			if n := v.getState(c); n != nil {
@@ -938,6 +963,8 @@ func (c *OpContext) lookup(x *Vertex, pos token.Pos, l Feature, flags combinedFl
 		}
 	} else {
 		if x.state != nil {
+			x.state.assertInitialized()
+
 			for _, e := range x.state.exprs {
 				if isCyclePlaceholder(e.err) {
 					hasCycle = true
@@ -948,7 +975,7 @@ func (c *OpContext) lookup(x *Vertex, pos token.Pos, l Feature, flags combinedFl
 		// As long as we have incomplete information, we cannot mark the
 		// inability to look up a field as "final", as it may resolve down the
 		// line.
-		permanent := x.status > conjuncts
+		permanent := x.status >= conjuncts
 		if m, _ := x.BaseValue.(*ListMarker); m != nil && !m.IsOpen {
 			permanent = true
 		}
@@ -1122,7 +1149,7 @@ func (c *OpContext) scalar(v Value) Value {
 	return v
 }
 
-var zero = &Num{K: NumKind}
+var zero = &Num{K: NumberKind}
 
 func (c *OpContext) Num(v Value, as interface{}) *Num {
 	v = Unwrap(v)
@@ -1131,7 +1158,7 @@ func (c *OpContext) Num(v Value, as interface{}) *Num {
 	}
 	x, ok := v.(*Num)
 	if !ok {
-		c.typeErrorAs(v, NumKind, as)
+		c.typeErrorAs(v, NumberKind, as)
 		return zero
 	}
 	return x
@@ -1209,7 +1236,7 @@ func (c *OpContext) ToBytes(v Value) []byte {
 
 // ToString returns the string value of a scalar value.
 func (c *OpContext) ToString(v Value) string {
-	return c.toStringValue(v, StringKind|NumKind|BytesKind|BoolKind, nil)
+	return c.toStringValue(v, StringKind|NumberKind|BytesKind|BoolKind, nil)
 
 }
 
@@ -1361,7 +1388,7 @@ func (c *OpContext) Str(x Node) string {
 	if c.Format == nil {
 		return fmt.Sprintf("%T", x)
 	}
-	return c.Format(x)
+	return c.Format(c.Runtime, x)
 }
 
 // NewList returns a new list for the given values.
