@@ -12,7 +12,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
-package validate
+package validate_test
 
 import (
 	"fmt"
@@ -24,6 +24,7 @@ import (
 	"cuelang.org/go/internal/core/adt"
 	"cuelang.org/go/internal/core/compile"
 	"cuelang.org/go/internal/core/eval"
+	"cuelang.org/go/internal/core/validate"
 	"cuelang.org/go/internal/cuetdtest"
 	"github.com/google/go-cmp/cmp"
 )
@@ -34,13 +35,13 @@ func TestValidate(t *testing.T) {
 		in     string
 		out    string
 		lookup string
-		cfg    *Config
+		cfg    *validate.Config
 
 		todo_v3 bool
 	}
 	testCases := []testCase{{
 		name: "no error, but not concrete, even with definition label",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		#foo: { use: string }
 		`,
@@ -48,13 +49,13 @@ func TestValidate(t *testing.T) {
 		out:    "incomplete\n#foo.use: incomplete value string:\n    test:2:16",
 	}, {
 		name: "definitions not considered for completeness",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		#foo: { use: string }
 		`,
 	}, {
 		name: "hidden fields not considered for completeness",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		_foo: { use: string }
 		`,
@@ -84,7 +85,7 @@ func TestValidate(t *testing.T) {
 		out: "eval\nx: conflicting values 2 and 1:\n    test:2:6\n    test:2:10",
 	}, {
 		name: "all errors",
-		cfg:  &Config{AllErrors: true},
+		cfg:  &validate.Config{AllErrors: true},
 		in: `
 		x: 1 & 2
 		y: 2 & 4
@@ -98,7 +99,7 @@ y: conflicting values 4 and 2:
     test:3:10`,
 	}, {
 		name: "incomplete",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		y: 2 + x
 		x: string
@@ -112,7 +113,7 @@ y: conflicting values 4 and 2:
 		`,
 	}, {
 		name: "allowed incomplete when disallowing cycles",
-		cfg:  &Config{DisallowCycles: true},
+		cfg:  &validate.Config{DisallowCycles: true},
 		in: `
 		y: string
 		x: y
@@ -122,7 +123,7 @@ y: conflicting values 4 and 2:
 		todo_v3: true,
 
 		name: "disallow cycle",
-		cfg:  &Config{DisallowCycles: true},
+		cfg:  &validate.Config{DisallowCycles: true},
 		in: `
 		y: x + 1
 		x: y - 1
@@ -133,7 +134,7 @@ y: conflicting values 4 and 2:
 		todo_v3: true,
 
 		name: "disallow cycle",
-		cfg:  &Config{DisallowCycles: true},
+		cfg:  &validate.Config{DisallowCycles: true},
 		in: `
 		a: b - 100
 		b: a + 100
@@ -141,7 +142,7 @@ y: conflicting values 4 and 2:
 		out: "cycle\ncycle error:\n    test:2:6",
 	}, {
 		name: "treat cycles as incomplete when not disallowing",
-		cfg:  &Config{},
+		cfg:  &validate.Config{},
 		in: `
 		y: x + 1
 		x: y - 1
@@ -150,7 +151,7 @@ y: conflicting values 4 and 2:
 		// Note: this is already handled by evaluation, as terminal errors
 		// are percolated up.
 		name: "catch most serious error",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		y: string
 		x: 1 & 2
@@ -158,13 +159,13 @@ y: conflicting values 4 and 2:
 		out: "eval\nx: conflicting values 2 and 1:\n    test:3:6\n    test:3:10",
 	}, {
 		name: "consider defaults for concreteness",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		x: *1 | 2
 		`,
 	}, {
 		name: "allow non-concrete in definitions in concrete mode",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		x: 2
 		#d: {
@@ -174,7 +175,7 @@ y: conflicting values 4 and 2:
 		`,
 	}, {
 		name: "pick up non-concrete value in default",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		x: null | *{
 			a: int
@@ -183,7 +184,7 @@ y: conflicting values 4 and 2:
 		out: "incomplete\nx.a: incomplete value int:\n    test:3:7",
 	}, {
 		name: "pick up non-concrete value in default",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 			x: null | *{
 				a: 1 | 2
@@ -195,7 +196,7 @@ y: conflicting values 4 and 2:
 		todo_v3: true,
 
 		name: "required field not present",
-		cfg:  &Config{Final: true},
+		cfg:  &validate.Config{Final: true},
 		in: `
 			Person: {
 				name!:  string
@@ -206,7 +207,7 @@ y: conflicting values 4 and 2:
 		out: "incomplete\nPerson.name: field is required but not present:\n    test:3:5",
 	}, {
 		name: "allow required fields in definitions",
-		cfg:  &Config{Concrete: true},
+		cfg:  &validate.Config{Concrete: true},
 		in: `
 		#Person: {
 			name!: string
@@ -245,7 +246,7 @@ y: conflicting values 4 and 2:
 			v = v.Lookup(adt.MakeIdentLabel(r, tc.lookup, "main"))
 		}
 
-		b := Validate(ctx, v, tc.cfg)
+		b := validate.Validate(ctx, v, tc.cfg)
 
 		w := &strings.Builder{}
 		if b != nil {
