@@ -31,6 +31,7 @@ import (
 	"strings"
 
 	"cuelabs.dev/go/oci/ociregistry"
+	"cuelabs.dev/go/oci/ociregistry/ociref"
 	"cuelang.org/go/internal/mod/semver"
 	digest "github.com/opencontainers/go-digest"
 	specs "github.com/opencontainers/image-spec/specs-go"
@@ -173,7 +174,7 @@ func (c *Client) GetModuleWithManifest(m module.Version, contents []byte, mediaT
 // sorted in semver order.
 // If m has a major version suffix, only versions with that major version will
 // be returned.
-func (c *Client) ModuleVersions(ctx context.Context, m string) ([]string, error) {
+func (c *Client) ModuleVersions(ctx context.Context, m string) (_req []string, _err0 error) {
 	mpath, major, hasMajor := module.SplitPathVersion(m)
 	if !hasMajor {
 		mpath = m
@@ -186,6 +187,12 @@ func (c *Client) ModuleVersions(ctx context.Context, m string) ([]string, error)
 		return nil, err
 	}
 	versions := []string{}
+	if !ociref.IsValidRepository(loc.Repository) {
+		// If it's not a valid repository, it can't be used in an OCI
+		// request, so return an empty slice rather than the
+		// "invalid OCI request" error that a registry can return.
+		return nil, nil
+	}
 	// Note: do not use c.repoName because that always expects
 	// a module path with a major version.
 	iter := loc.Registry.Tags(ctx, loc.Repository, "")
@@ -343,17 +350,11 @@ func checkModFile(m module.Version, f *zip.File) ([]byte, *modfile.File, error)
 	if err != nil {
 		return nil, nil, err
 	}
-	if mf.Module != m.Path() {
-		return nil, nil, fmt.Errorf("module path %q found in %s does not match module path being published %q", mf.Module, f.Name, m.Path())
-	}
-	_, major, ok := module.SplitPathVersion(mf.Module)
-	if !ok {
-		// Note: can't happen because we already know that mf.Module is the same
-		// as m.Path which is a valid module path.
-		return nil, nil, fmt.Errorf("invalid module path %q", mf.Module)
+	if mf.QualifiedModule() != m.Path() {
+		return nil, nil, fmt.Errorf("module path %q found in %s does not match module path being published %q", mf.QualifiedModule(), f.Name, m.Path())
 	}
 	wantMajor := semver.Major(m.Version())
-	if major != wantMajor {
+	if major := mf.MajorVersion(); major != wantMajor {
 		// This can't actually happen because the zip checker checks the major version
 		// that's being published to, so the above path check also implicitly checks that.
 		return nil, nil, fmt.Errorf("major version %q found in %s does not match version being published %q", major, f.Name, m.Version())
